//
//
//

module compress;

@safe:

import std.stdio;

struct mynode{
	ubyte value;
	int len;
	uint code;
}

struct mybyte{
	ubyte letter;
	int count;
}

mynode[256] node;

private void showbits(uint num){
	showbits(num,32);
}

private void showbits(uint num,int ln){
	for(int i=0;i<ln;i++){
		uint mark=(0x80000000 & 0xffffffff)>>i;
		if ((num & mark)==mark){
			write("1");
		}else{
			write("0");
		}
	}
}

void gencodes(bool dbg){
	int l=0;
	foreach(i;0..16){
		foreach(j;0..16){
			int test=0;
			node[l].len=2+i+j;
		
			// Fill byte with 1's on left
			foreach(s;0..i){
				test=test | (0x80000000 & 0xffffffff)>>s;
			}
		
			test=test | (0x40000000 & 0xffffffff)>>(i+j);
	
			node[l].code=test;
		
			l++;
		}
	}
	
	//
	// Sort code lengths small>big
	//
	mynode tnode;
	
	while(true){
		auto done=true;
		foreach(i;0..255){
			if (node[i].len>node[i+1].len){ 
				tnode=node[i];	
				node[i]=node[i+1];
				node[i+1]=tnode;
				done=false;
			}
		}
		if (done==true){break;}
	}
	
	// Show codes
	if (dbg==true){
		foreach(i;0..256) {
			write(i,": ");
			write("\t",node[i].len);
			write("\t");
			showbits(node[i].code);
			writeln;
		}
	}
	
}


private void showcodes(){
	foreach(i;0..5) {
		write(i,": ");
		write(": ",cast(char)node[i].value);
		write("\t",node[i].len);
		write("\t");
		showbits(node[i].code,node[i].len);
		writeln;
	}
}

/*
private int codematch(uint input){
	auto i=0;
	foreach(n;node){
		if (n.code==input){return i;}
		i++;
	}
	return -1;
}
*/

bool dbg=false;
mybyte[] bytelist;

bool go_compress(ubyte[] input,ref ubyte[] output){

	int[256] count;
	
	// Count the frequency of the letters in the text
	foreach(byte c;input){
		ubyte num=cast(ubyte)(c);
		// writeln(">",num);
		if (num<256){
			count[num]++;
		}
	}
	
	// Make a list of the byte / frequency so we can sort them
	// mybyte[] bytelist;
	auto v=0;
	foreach(c;count){
		if (c>0){
			mybyte tmp;
			tmp.letter=cast(byte)v;
			tmp.count=c;
			bytelist~=tmp;
		}
		v++;
	}
	
	// Sort them so highest count gets smallest byte
	mybyte tbyte;
	while(true){
		auto done=true;
		foreach(i;0..bytelist.length-1){
			if (bytelist[i].count<bytelist[i+1].count){ 
				tbyte=bytelist[i];	
				bytelist[i]=bytelist[i+1];
				bytelist[i+1]=tbyte;
				done=false;
			}
		}
		if (done==true){break;}
	}
	
	// Show approx size
	auto size_in_bits=0;
	auto y=0;
	foreach(b;bytelist){
		
		node[y].value=b.letter;
		
		// write(b.letter);
		// write(" -  ",b.count," x ",node[y].len);
		// writeln;
		
		size_in_bits+=(node[y].len*b.count);
		y++;
	}
	
	writeln("Size:\t",input.length," bytes");
	writeln("Comp:\t",(size_in_bits/8)," bytes (aprrox)");
	
	// See if compression worth it...
	if ((size_in_bits/8)>(input.length-20)){
		return false;
	}
	
	// Write compressed txt to array
	// auto dbg=false;
	auto bytepos=0;
	
	char outbyte=0;
	
	foreach(ch;input){
		// Search through byte list to get code position
		auto cz=-1;
		auto i=0;
		foreach(n;node){
			if (n.value==ch){
				cz=i;
				break;
			}
			i++;
		}
		
		if (dbg==true){
			if (ch<32)
			{write("?");}
			else
			{writef("%c",cast(char)ch);}
			
			writef(" (%d) \t> node: ",ch);
			showbits(node[cz].code,node[cz].len);
			//prn(node[cz].code,node[cz].len);
			writeln; 			
		}
		
		if (cz<0){
			writeln("Code ",ch," not found!!");
		}else{
			uint bit=0x80000000;
			
			foreach(j;0..node[cz].len){
				if ( (node[cz].code & bit)==bit){	
					outbyte+=(0x80 & 0xff)>>bytepos; 
				}
						
				bit=(bit & 0xffffffff)>>1;

				bytepos++;
				if (bytepos==8){
					output~=outbyte;
					// showbits(outbyte,8);
					bytepos=0;
					outbyte=0;
				}	
			}
		}
	}
	if (bytepos>0){
		output~=outbyte;
		// showbits(outbyte,8);
	}	
	
	writeln;
	// Compressed string
	if (dbg==true){
		writeln("Compressed bit stream:");
		foreach(ch;output){
			uint tmp=(ch &0xffffffff)<<24;
			showbits(tmp,8);
		}
		writeln;
	}
	
	return true;
}

bool go_decompress(ubyte[] input,ref ubyte[] output){
	int l;
	uint code;
	auto size=0;
	
	// Assuming that output is at least as big as the input
	output.length=input.length;
	
	// Saves doing lots of shifts
	ubyte[8] va=[0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01];
	
	// Make our Associative Array
	ubyte[int] map;
	foreach(n;node){
		map[n.code]=n.value;
	}
	
	foreach(ubyte ch;input){
		foreach(i;0..8){
			if ((ch & va[i])==va[i] ){
				code+=(0x80000000>>l);
			}
			
			l++;			
			if (code in map){
				output[size]=map[code];
				code=0;
				l=0;
				size++;
				if (size==output.length){
					output.length+=1024;
				}
			}
		}
	}
	output.length=size;
	return true;
}
/*
bool slow_decompress(ubyte[] input,ref ubyte[] output){
	int l;
	uint code;
	
	foreach(ubyte ch;input){
		ubyte va=cast(ubyte)0x80;
		foreach(i;0..8){
			if ((ch & va)==va ){
				code+=(0x80000000>>l);
			}
			va=(va & 0xff)>>1;
			
			l++;
			if (l>1){			
				auto outcode=codematch(code);
				if (outcode>=0){
					code=0;
					l=0;
					output~=node[outcode].value;
				}
			}
		}
	}
	return true;
}
*/